import os
import functools
import tempfile
import pkgutil

import boto3
import botocore
import click
import humanize

import s3sup.catalogue
import s3sup.fileprepper
import s3sup.rules
import s3sup.utils


def load_skeleton_s3sup_toml():
    return pkgutil.get_data(__package__, 'skeleton.s3sup.toml')


class Project:

    def __init__(self, local_project_root, dryrun=False,
                 preserve_deleted_files=False, verbose=True):
        self.dryrun = dryrun
        self.verbose = verbose
        self.local_project_root = local_project_root
        try:
            self.rules = s3sup.rules.load_rules(os.path.join(
                local_project_root, 's3sup.toml'))
        except FileNotFoundError:
            error_text = (
                '\n{0} not an s3sup project directory (no s3sup.toml found). '
                'Either:\n'
                ' * Change to an s3sup project directory before running.\n'
                ' * Supply project directory using -p/--projectdir.\n'
                ' * Create a new s3sup project direction using "s3sup init".'
            ).format(os.path.abspath(local_project_root))
            raise click.FileError(
                os.path.join(local_project_root, 's3sup.toml'),
                hint=error_text)

        self._preserve_deleted_files = preserve_deleted_files
        try:
            self._preserve_deleted_files = (
                preserve_deleted_files or self.rules['preserve_deleted_files'])
        except KeyError:
            pass

        self._fp_cache = {}
        self.local_preflight_checks()

    def _boto_bucket(self):
        s = boto3.session.Session()
        res_args = {}
        try:
            res_args['region_name'] = self.rules['aws']['region_name']
        except KeyError:
            pass
        try:
            res_args['endpoint_url'] = self.rules['aws']['s3_endpoint_url']
        except KeyError:
            pass
        r = s.resource(service_name='s3', **res_args)
        b = r.Bucket(self.rules['aws']['s3_bucket_name'])
        return r, b

    def file_prepper_wrapped(self, path):
        try:
            return self._fp_cache[path]
        except KeyError:
            self._fp_cache[path] = s3sup.fileprepper.FilePrepper(
                self.local_project_root, path, self.rules)
        return self._fp_cache[path]

    def _local_fs_path(self, rel_path):
        return os.path.join(self.local_project_root, rel_path)

    def local_preflight_checks(self):
        """
        Can't stop things going wrong during the upload, but having a good poke
        round catches most problems.
        """
        return True

    def remote_preflight_checks(self):
        """
        Can't stop things going wrong during the upload, but having a good poke
        round catches most problems.
        """
        rmt_cat_fp = self.file_prepper_wrapped('.s3sup.write_test')
        rsrc, b = self._boto_bucket()
        o = b.Object(rmt_cat_fp.s3_path())
        try:
            o.put(Body='Can s3sup write to bucket?', ACL='private')
        except rsrc.meta.client.exceptions.NoSuchBucket:
            raise click.ClickException('S3 bucket does not exist: {0}'.format(
                self.rules['aws']['s3_bucket_name']))
        o.delete()

    @functools.lru_cache(maxsize=8)
    def local_catalogue(self):
        local_cat = s3sup.catalogue.Catalogue(
            preserve_deleted_files=self._preserve_deleted_files)
        for root, dirs, files in os.walk(self.local_project_root):
            for f in files:
                if f == 's3sup.toml':
                    continue
                abs_path = os.path.join(root, f)
                rel_path = os.path.relpath(
                    abs_path, start=self.local_project_root)
                fp = self.file_prepper_wrapped(rel_path)
                local_cat.add_file(
                    rel_path, fp.content_hash(), fp.attributes_hash())
        return local_cat

    @functools.lru_cache(maxsize=8)
    def get_remote_catalogue(self):
        remote_cat = s3sup.catalogue.Catalogue(
            preserve_deleted_files=self._preserve_deleted_files)

        _, b = self._boto_bucket()
        old_cat_fp = self.file_prepper_wrapped('.s3sup.catalogue.csv')
        old_f = b.Object(old_cat_fp.s3_path())

        new_cat_fp = self.file_prepper_wrapped('.s3sup.cat')
        new_f = b.Object(new_cat_fp.s3_path())

        hndl, tmpp = tempfile.mkstemp()
        os.close(hndl)
        try:
            new_f.download_file(tmpp)
            remote_cat.from_sqlite(tmpp)
        except botocore.exceptions.NoCredentialsError:
            raise click.UsageError(
                'Cannot find AWS credentials.\n -> Configure AWS credentials '
                ' using any method that the underlying boto3 library supports:'
                '\n -> https://boto3.amazonaws.com/v1/documentation/'
                'api/latest/guide/configuration.html')
        except botocore.exceptions.ClientError:
            if self.verbose:
                click.echo(
                    ('Could not find SQLite based remote catalogue on S3 '
                     '(expected at {0}).').format(new_cat_fp.s3_path()))
            try:
                old_f.download_file(tmpp)
                remote_cat.from_csv(tmpp)
                click.echo(click.style((
                    'WARNING: After the next s3sup push, do not attempt to '
                    'use older versions of s3sup (0.3.0 or below) with this '
                    'project, as they will no longer be able to read the '
                    'remote catalogue.'), fg='blue'))
            except botocore.exceptions.ClientError:
                if self.verbose:
                    click.echo(
                        ('Could not find older CSV based remote catalogue on '
                         'S3 either (expected at {0}). This indicates the '
                         'project has never been pushed to S3 before.').format(
                            old_cat_fp.s3_path()))
                pass
            pass
        os.remove(tmpp)
        return remote_cat

    def write_remote_catalogue(self, catalogue):
        hndl, tmpp = tempfile.mkstemp()
        os.close(hndl)
        catalogue.to_sqlite(tmpp)
        rmt_cat_fp = self.file_prepper_wrapped('.s3sup.cat')
        _, b = self._boto_bucket()
        o = b.Object(rmt_cat_fp.s3_path())
        with open(tmpp, 'rb') as lf:
            o.put(Body=lf, ACL='private')
        os.remove(tmpp)

        # Deliberately break older s3sup clients <= 0.3.0.
        # This file even needs uploading even for projects that have never used
        # the old format, just in-case an old version of s3sup is used on it
        # in the future (perhaps if part of a CI/CD system is used).
        old_rmt_cat_fp = self.file_prepper_wrapped('.s3sup.catalogue.csv')
        the_breaker = (
            b'\xF9\xF9This is a deliberately corrupt old version of the s3sup '
            b'catalogue format It is not used any more and this file is only '
            b'here to cause s3sup clients <= 0.3.0 to fail, rather than have '
            b'them try to upload everything again.')
        b.Object(old_rmt_cat_fp.s3_path()).put(Body=the_breaker, ACL='private')

    def calculate_diff(self):
        local_cat = self.local_catalogue()
        remote_cat = self.get_remote_catalogue()
        diff, new_remote_cat = local_cat.diff_dict(remote_cat)
        return (diff, new_remote_cat)

    def sync(self):
        self.remote_preflight_checks()
        diff, new_remote_cat = self.calculate_diff()
        changes = s3sup.catalogue.change_list(diff)
        changes_with_prep = [
            (cr, p, self.file_prepper_wrapped(p)) for cr, p in changes]

        if len(changes) <= 0:
            return changes

        if self.dryrun:
            click.echo(click.style(
                'Not making any changes as this is a dry run.', fg='blue'))
            return changes

        _, b = self._boto_bucket()

        def display_current(item):
            if item is None:
                return ''
            cr, p, fp = item
            crs = s3sup.catalogue.CR_STYLES[cr]
            change_symbol = click.style(
                '{symbol}'.format(symbol=getattr(crs, 'symbol')),
                fg=getattr(crs, 'colour'))

            cur = ' {0} {1}'.format(change_symbol, fp.s3_path())
            if (cr == s3sup.catalogue.ChangeReason['NEW_FILE'] or
                    cr == s3sup.catalogue.ChangeReason['CONTENT_CHANGED']):
                cur += ' ({0})'.format(humanize.naturalsize(fp.size()))
            return cur

        with click.progressbar(changes_with_prep, label='Syncing to S3',
                               item_show_func=display_current) as bar:
            for cr, p, fp in bar:
                o = b.Object(fp.s3_path())
                if cr == s3sup.catalogue.ChangeReason['NEW_FILE']:
                    with fp.content_fileobj() as lf:
                        o.put(Body=lf, **fp.attributes_as_boto_args())
                elif cr == s3sup.catalogue.ChangeReason['CONTENT_CHANGED']:
                    with fp.content_fileobj() as lf:
                        o.put(Body=lf, **fp.attributes_as_boto_args())
                elif cr == s3sup.catalogue.ChangeReason['ATTRIBUTES_CHANGED']:
                    o.copy_from(
                        CopySource={
                            'Bucket': self.rules['aws']['s3_bucket_name'],
                            'Key': fp.s3_path()},
                        MetadataDirective='REPLACE',
                        TaggingDirective='REPLACE',
                        **fp.attributes_as_boto_args())
                elif cr == s3sup.catalogue.ChangeReason['DELETED']:
                    o.delete()
                else:
                    raise Exception('Unknown ChangeReason: {0}'.format(cr))

        self.write_remote_catalogue(new_remote_cat)
        return changes

    def print_summary(self):
        lcl_dir = click.format_filename(self.local_project_root)
        if lcl_dir == '.':
            lcl_dir += ' (current dir)'

        s3p = 's3://{0}/'.format(self.rules['aws']['s3_bucket_name'])
        try:
            s3p += self.rules['aws']['s3_project_root'].lstrip('/').rstrip('/')
        except KeyError:
            pass

        to_print = {
            'Local project dir': lcl_dir,
            'AWS region': self.rules['aws']['region_name'],
            'S3 bucket': s3p
        }
        s3sup.utils.pprint_h1('PROJECT INFORMATION')
        s3sup.utils.pprint_dict(to_print)



import csv
import contextlib
import gzip
import sqlite3
import shutil
import tempfile
import enum
import collections
import copy
import click
import inflect


class ChangeReason(enum.Enum):
    NEW_FILE = 1
    CONTENT_CHANGED = 2
    ATTRIBUTES_CHANGED = 3
    DELETED = 4
    DELETED_PROTECTED = 5
    NO_CHANGE = 6


CR_STYLE = collections.namedtuple(
    'CR_STYLE', ['colour', 'symbol', 'shortreason', 'longreason'])

CR_STYLES = {
    ChangeReason['NEW_FILE']: CR_STYLE('yellow', '+', 'new', 'new'),
    ChangeReason['CONTENT_CHANGED']: CR_STYLE(
        'blue', '*', 'changed', 'content changed'),
    ChangeReason['ATTRIBUTES_CHANGED']: CR_STYLE(
        'cyan', '~', 'attrs', 'attributes changed'),
    ChangeReason['DELETED']: CR_STYLE('red', '-', 'delete', 'deleted'),
    ChangeReason['DELETED_PROTECTED']: CR_STYLE(
        'green', '^', 'deleteprotected', 'deleted but protected'),
    ChangeReason['NO_CHANGE']: CR_STYLE('green', '^', 'unchanged', 'unchanged')
}

MAX_DB_SCHEMA_VERSION = 2


@contextlib.contextmanager
def load_gzipped_sqlite(path):
    with tempfile.NamedTemporaryFile() as out_f:
        with gzip.open(path, 'rb') as in_f:
            shutil.copyfileobj(in_f, out_f)
        c = sqlite3.connect(out_f.name)
        c.row_factory = sqlite3.Row
        yield c
        c.close()


@contextlib.contextmanager
def write_gzipped_sqlite(path):
    with tempfile.NamedTemporaryFile() as in_f:
        c = sqlite3.connect(in_f.name)
        c.row_factory = sqlite3.Row
        yield c
        c.commit()
        c.close()
        with gzip.open(path, 'wb') as out_f:
            shutil.copyfileobj(in_f, out_f)


class Catalogue:

    def __init__(self, preserve_deleted_files=False):
        self._c = {}
        self._preserve_deleted_files = preserve_deleted_files

    def add_file(self, path: str, content_hash: str, attributes_hash: str):
        self._c[path] = (str(content_hash), str(attributes_hash))
        return self

    def to_dict(self):
        return {k: self._c[k] for k in sorted(self._c.keys())}

    def from_csv(self, path: str):
        with open(path, 'rt', newline='') as f:
            reader = csv.reader(f)
            next(reader)  # Skip CSV header
            for path, content_hash, attributes_hash in csv.reader(f):
                self.add_file(path, content_hash, attributes_hash)

    def from_sqlite(self, path: str):
        with load_gzipped_sqlite(path) as c:
            schema_version = c.execute('PRAGMA user_version').fetchone()[0]
            if schema_version > MAX_DB_SCHEMA_VERSION:
                raise click.ClickException((
                    'Upgrade to latest s3sup to continue. The s3sup version'
                    'last used to push this project to S3 was newer than the '
                    'installed version. The newer remote catalogue format is '
                    'not readable by older s3sup version. Catalogue schema is '
                    'version {0}, this s3sup only supports catalogue schema '
                    'up to version {1}.').format(
                        schema_version, MAX_DB_SCHEMA_VERSION))
            # Handle migrations here
            for row in c.execute('SELECT * FROM files'):
                self.add_file(
                    row['path'], row['content_hash'], row['attributes_hash'])

    def to_sqlite(self, path: str):
        with write_gzipped_sqlite(path) as c:
            c.execute('PRAGMA user_version = {v:d}'.format(
                v=MAX_DB_SCHEMA_VERSION))
            c.execute('''CREATE TABLE files (
                path TEXT,
                content_hash TEXT,
                attributes_hash TEXT)''')
            c.executemany(
                'INSERT INTO files VALUES (?, ?, ?)',
                [(path, ch, ah) for path, (ch, ah) in self.to_dict().items()])

    def diff_dict(self, remote_catalogue):
        lcl = self.to_dict()
        rmt = remote_catalogue.to_dict()
        # TODO: Not a great use of memory, see if there's a better way.
        new_rmt = copy.deepcopy(self)

        changes = {
            'num_changes': 0,
            'upload': {
                'new_files': [],
                'content_changed': [],
                'attributes_changed': []
            },
            'delete': [],
            'delete_protected': [],
            'unchanged': []
        }

        for path, (content_hash, attributes_hash) in rmt.items():
            if path not in lcl:
                if self._preserve_deleted_files:
                    changes['delete_protected'].append(path)
                    # Deleted but protected files are the only ones that need
                    # to be maintained in remote
                    new_rmt.add_file(path, content_hash, attributes_hash)
                else:
                    changes['delete'].append(path)

        for path, (content_hash, attributes_hash) in lcl.items():
            try:
                r_content_hash, r_attributes_hash = rmt[path]
                if content_hash != r_content_hash:
                    changes['upload']['content_changed'].append(path)
                elif attributes_hash != r_attributes_hash:
                    changes['upload']['attributes_changed'].append(path)
                else:
                    changes['unchanged'].append(path)
            except KeyError:
                changes['upload']['new_files'].append(path)

        changes['num_changes'] = (
            len(changes['delete'])
            + len(changes['upload']['new_files'])
            + len(changes['upload']['content_changed'])
            + len(changes['upload']['attributes_changed'])
        )
        return changes, new_rmt


def print_diff_summary(dd, verbose=False):
    ie = inflect.engine()
    nc = dd['num_changes']
    if nc <= 0:
        click.echo('No local changes to be synced.')
        return

    if verbose:
        click.echo('Local changes to be synced to S3:')
    else:
        click.echo('Summary of local changes to be synced to S3:')

    def _p(files, change_reason):
        num_files = len(files)
        if num_files <= 0:
            return
        crs = CR_STYLES[ChangeReason[change_reason]]
        cr_prefix = click.style(
            '{symbol} {changereason}'.format(
                symbol=getattr(crs, 'symbol'),
                changereason=getattr(crs, 'longreason')),
            fg=getattr(crs, 'colour'))
        files_txt = '{0} {1}'.format(num_files, ie.plural('file', num_files))
        click.echo(' {cr_prefix}: {files_txt}'.format(
            cr_prefix=cr_prefix,
            files_txt=files_txt))

        if not verbose or change_reason == 'NO_CHANGE':
            return

        change_symbol = click.style(
            '{symbol}'.format(symbol=getattr(crs, 'symbol')),
            fg=getattr(crs, 'colour'))

        for p in files:
            click.echo('   {0} {1}'.format(change_symbol, p))

    _p(dd['upload']['new_files'], 'NEW_FILE')
    _p(dd['upload']['content_changed'], 'CONTENT_CHANGED')
    _p(dd['upload']['attributes_changed'], 'ATTRIBUTES_CHANGED')
    _p(dd['delete'], 'DELETED')
    _p(dd['delete_protected'], 'DELETED_PROTECTED')
    _p(dd['unchanged'], 'NO_CHANGE')


def _order_for_upload(path_names):
    """
    Prevent HTML files referencing static assets (stylesheets/scripts/images)
    before they exist on S3. Achieved by applying a crude but effective
    ordering to the upload sequence to ensure HTML files are uploaded after all
    other assets.
    """
    html, css, js, others = [], [], [], []
    for p in path_names:
        pl = p.lower()
        if pl.endswith(('.html', '.htm', '.xhtml')):
            html.append(p)
        elif pl.endswith(('.css')):
            css.append(p)
        elif pl.endswith(('.js')):
            js.append(p)
        else:
            others.append(p)

    def _srt(g):
        return sorted(sorted(g), key=lambda x: x.count('/'), reverse=True)

    ordered = _srt(others) + _srt(css) + _srt(js) + _srt(html)
    assert len(ordered) == len(path_names)
    return ordered


def change_list(diff):
    """
    Paths that need changes made on S3, along with the reason why.

    Changes should be made in the order returned, in an attempt to prevent the
    temporary period where there may be incorrect links between HTML files,
    referencing static assets that haven't been uploaded yet.
    """
    # Attribute changes first, less risk of content changes.
    dl = [(ChangeReason.ATTRIBUTES_CHANGED, p)
          for p in diff['upload']['attributes_changed']]
    dl += [(ChangeReason.NEW_FILE, p)
           for p in _order_for_upload(diff['upload']['new_files'])]
    dl += [(ChangeReason.CONTENT_CHANGED, p)
           for p in _order_for_upload(diff['upload']['content_changed'])]
    dl += [(ChangeReason.DELETED, p)
           for p in diff['delete']]
    return dl



import os
import sys
import functools
import click
import pathlib

sys.path.insert(
    0, os.path.abspath(os.path.join(os.path.dirname(__file__), '../..')))
import s3sup.project  # noqa: E402
import s3sup.catalogue  # noqa: E402


def common_options(f):
    """
    Common command line options used by ALL s3sup commands
    """
    options = [
        click.option(
            '-p', '--projectdir', default='.',
            help=('Specify local s3sup static site directory (containing '
                  's3sup.toml). By default the current directory is used.')),
        click.option(
            '-v', '--verbose', is_flag=True,
            help='Output more informational messages than normal.')
    ]
    return functools.reduce(lambda x, opt: opt(x), options, f)


def options_for_remotes(f):
    """
    Command line options only used by s3sup commands that interact with S3.
    E.g. push and status.
    """
    options = [
        click.option(
            '-d', '--dryrun', is_flag=True,
            help='Simulate changes to be made. Do not modify files on S3.'),
        click.option(
            '-n', '--nodelete', is_flag=True,
            help=('Do not delete any files on S3, add/modify operations only. '
                  'Alternatively set "preserve_deleted_files" '
                  'in s3sup.toml.')),
    ]
    return functools.reduce(lambda x, opt: opt(x), options, f)


class AliasedGroup(click.Group):
    """
    Provides shortcut commands, allowing 'st' to be typed for 'status'.
    Skeleton class taken straight from the click documentation.
    """

    def get_command(self, ctx, cmd_name):
        rv = click.Group.get_command(self, ctx, cmd_name)
        if rv is not None:
            return rv
        commands = self.list_commands(ctx)
        aliases = {
            'upload': 'push',
            'sync': 'push'
        }
        all_commands = commands + [alias for alias in aliases]
        matches = [x for x in all_commands if x.startswith(cmd_name)]
        if not matches:
            return None
        elif len(matches) == 1:
            try:
                match = aliases[matches[0]]
            except KeyError:
                match = matches[0]
            return click.Group.get_command(self, ctx, match)
        ctx.fail('Multiple commands matched "{0}": {1}'.format(
            cmd_name, ', '.join(sorted(matches))))


# Would rather add options to click.group(), but this doesn't work as intended:
# https://stackoverflow.com/questions/52144383/
@click.command(cls=AliasedGroup)
def cli():
    """
    s3sup - Amazon S3 static site uploader
    """
    pass


@cli.command()
@common_options
def init(projectdir, verbose):
    """
    Create a skeleton s3sup.toml in the current directory.
    """
    # Check file doesn't already exist
    cf = pathlib.Path(os.path.join(projectdir, 's3sup.toml'))
    if cf.exists():
        raise click.FileError(
            's3sup.toml', hint='s3sup configuration file already exists')
    cf.write_bytes(s3sup.project.load_skeleton_s3sup_toml())
    click.echo('Skeleton configuration file created: {0}'.format(
        cf.absolute()))


@cli.command()
@common_options
@options_for_remotes
def status(projectdir, verbose, dryrun, nodelete):
    """
    Show S3 changes that will be made on next push.
    """
    click.echo('S3 site uploader. Using:')
    p = s3sup.project.Project(
        projectdir, dryrun=dryrun, preserve_deleted_files=nodelete,
        verbose=verbose)
    if verbose or projectdir != '.':
        click.echo(' * Local project directory: {0}'.format(projectdir))

    try:
        s3_root = p.rules['aws']['s3_project_root'].strip()
    except KeyError:
        s3_root = '/'
    s3_root = '/' if s3_root == '' else s3_root

    if s3_root != '/':
        s3_root = '/{0}/'.format(s3_root.lstrip('/').rstrip('/'))

    if verbose:
        click.echo(' * AWS region: {0}'.format(p.rules['aws']['region_name']))

    click.echo(' * S3 bucket: {0}{1}'.format(
        p.rules['aws']['s3_bucket_name'], s3_root))

    click.echo('')
    diff, _ = p.calculate_diff()
    s3sup.catalogue.print_diff_summary(diff, verbose=True)


@cli.command()
@click.argument('local_file', nargs=-1)
@common_options
def inspect(local_file, projectdir, verbose):
    """
    Show calculated metadata for individual files.
    """
    p = s3sup.project.Project(projectdir, dryrun=True, verbose=verbose)
    p.print_summary()
    for f in local_file:
        click.echo()
        try:
            fp = p.file_prepper_wrapped(f)
            fp.print_summary()
        except FileNotFoundError:
            msg = 'Could not open: {0}'.format(click.format_filename(f))
            click.echo(click.style(msg, fg='red'), err=True)


@cli.command()
@common_options
@options_for_remotes
def push(projectdir, verbose, dryrun, nodelete):
    """
    Synchronise local static site to S3.

    Use --dryrun to test behaviour without changes actually being made to S3.
    Or use "s3sup status".

    This command has two other aliases: upload or sync.
    """
    p = s3sup.project.Project(
        projectdir, dryrun=dryrun, preserve_deleted_files=nodelete,
        verbose=verbose)
    diff, _ = p.calculate_diff()
    s3sup.catalogue.print_diff_summary(diff, verbose=verbose)
    p.sync()
    click.echo(click.style('Done!', fg='green'))


if __name__ == '__main__':
    cli()



from setuptools import setup, find_packages
import os

here = os.path.abspath(os.path.dirname(__file__))

with open(os.path.join(here, 'README.md'), encoding='utf-8') as f:
    # Filter out demo GIFs from the pypi version
    long_description = ''.join(
        l for l in f.readlines()
        if '<img src' not in l and ' demo ' not in l)

setup(
    name='s3sup',
    version='0.5.0',
    description='Static site uploader for Amazon S3',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/awooldrige/s3sup',
    author='Alistair Wooldrige',
    author_email='s3sup@woolie.co.uk',

    # For a list of valid classifiers, see https://pypi.org/classifiers/
    classifiers=[  # Optional
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7'
    ],
    keywords='s3sup AWS S3 static',
    packages=find_packages(exclude=['contrib', 'docs', 'tests']),
    python_requires='>=3, <4',
    install_requires=[
        'boto3>=1,<2',
        'click>=7,<9',
        'humanize>=0.4,<1',
        'inflect>=2,<3',
        'jsonschema>=2,<4',
        'toml>=0.9,<1'
    ],
    include_package_data=True,
    extras_require={
        'test': ['flake8', 'moto'],
    },
    entry_points={
        'console_scripts': [
            's3sup=s3sup.scripts.s3sup:cli',
        ],
    },
    project_urls={
        'Bug Reports': 'https://github.com/awooldrige/s3sup/issues',
        'Source': 'https://github.com/awooldrige/s3sup/'
    }
)



import click


def pprint_h1(text):
    click.echo('*'*60)
    click.echo('* {0}'.format(text))
    click.echo('*'*60)


def pprint_h2(text):
    click.echo('* {0}'.format(text))
    click.echo('*'*60)


def pprint_h3(text):
    click.echo('{0}'.format(text))
    click.echo('*'*len(text))


def pprint_dict(d, level=0, level_name=None):
    try:
        iterator = d.items()
        if level_name is not None:
            click.echo('  '*level + '- {0}:'.format(level_name))
        for k, v in iterator:
            pprint_dict(v, level=level+1, level_name=k)
    except AttributeError:
        click.echo('  '*level + '- {0}: {1}'.format(
            level_name,
            click.style(str(d), fg='green')))



import functools
import pathlib
import pickle
import hashlib
import mimetypes
import collections

import click
import humanize

import s3sup.rules


ADDITIONAL_MIMETYPES = {
    '.toml': 'application/toml',
    '.woff': 'font/woff',
    '.woff2': 'font/woff2'
}
TEXT_BASED_MIMETYPES = {
    'text/css',
    'text/csv',
    'text/html',
    'text/javascript',
    'text/ecmascript',
    'application/json',
    'application/toml',
    'application/javascript',
    'application/ecmascript',
    'application/rtf',
    'application/x-sh',
    'application/x-c',
    'application/x-latex',
    'image/svg+xml',
    'text/plain',
    'text/yaml',
    'application/yaml',
    'application/x-yaml',
    'application/xhtml+xml',
    'application/xml',
    'text/xml',
    'application/vnd.mozilla.xul+xml'
}
DEFAULT_CHARSET_MIMETYPES = TEXT_BASED_MIMETYPES
HASH_READ_BLOCK = 65536


class FilePrepper:

    def __init__(self, project_root, path, rules):
        self.project_root = project_root
        self.path = path

        self.path_proj = pathlib.Path(project_root)

        # Relative to project root. E.g. 'index.html'.
        self.path_local_rel = pathlib.Path(path)

        # Absolute path. E.g. '/home/jsmith/proj_1/index.html'
        self.path_local_abs = self.path_proj.joinpath(
            self.path_local_rel).resolve()

        self.rules = rules
        self.path_directives = s3sup.rules.directives_for_path(
            self.path, self.rules)

    @functools.lru_cache(maxsize=None)
    def attributes(self):
        # Defaults
        attrs = {
            'ACL': 'public-read',
            'Cache-Control': 'max-age=10'
        }
        fext = pathlib.Path(self.path).suffix
        mime_type, encoding = mimetypes.guess_type(self.path)
        try:
            mime_type = ADDITIONAL_MIMETYPES[fext]
        except KeyError:
            pass
        try:
            mime_type = self.rules['mimetype_overrides'][fext]
        except KeyError:
            pass

        charset_mimetypes = DEFAULT_CHARSET_MIMETYPES
        try:
            charset_mimetypes = self.rules['charset_mimetypes']
        except KeyError:
            pass

        charset = None
        if mime_type in charset_mimetypes:
            charset = 'utf-8'
            try:
                charset = self.rules['charset']
            except KeyError:
                pass
        try:
            charset = self.path_directives['charset']
        except KeyError:
            pass

        content_type = 'application/octet-stream'
        if mime_type is not None:
            content_type = mime_type
            if charset is not None:
                content_type += '; charset={0}'.format(charset)
        try:
            content_type = self.path_directives['Content-Type']
        except KeyError:
            pass
        attrs['Content-Type'] = content_type

        if encoding:
            attrs['Content-Encoding'] = encoding

        # conf_name: boto_name
        directives_to_transfer = {
            'ACL',
            'Cache-Control',
            'Content-Disposition',
            'Content-Language',
            'Content-Encoding',
            'S3Metadata',
            'StorageClass',
            'WebsiteRedirectLocation'
        }
        for dctv in directives_to_transfer:
            try:
                attrs[dctv] = self.path_directives[dctv]
            except KeyError:
                continue
        # Set as a sorted dict
        # self.attributes() = {k: attrs[k] for k in sorted(attrs)}
        return collections.OrderedDict(
            sorted(attrs.items(), key=lambda t: t[0]))

    def attributes_as_boto_args(self):
        key_map = {
            'ACL': 'ACL',
            'StorageClass': 'StorageClass',
            'WebsiteRedirectLocation': 'WebsiteRedirectLocation',
            'Cache-Control': 'CacheControl',
            'Content-Disposition': 'ContentDisposition',
            'Content-Type': 'ContentType',
            'Content-Encoding': 'ContentEncoding',
            'Content-Language': 'ContentLanguage',
            'S3Metadata': 'Metadata'
        }
        return {key_map[k]: v for k, v in self.attributes().items()}

    def s3_path(self):
        try:
            root = self.rules['aws']['s3_project_root']
        except KeyError:
            root = ''
        root = root.strip().strip('/')
        path = self.path_local_rel.as_posix()
        if len(root) > 0:
            return '{0}/{1}'.format(root, path)
        return path

    def content_fileobj(self):
        return self.path_local_abs.open('rb')

    def size(self):
        """Size of local file in bytes"""
        return self.path_local_abs.stat().st_size

    @functools.lru_cache(maxsize=None)
    def content_hash(self):
        sha = hashlib.sha256()
        with self.content_fileobj() as f_in:
            fbuf = f_in.read(HASH_READ_BLOCK)
            while len(fbuf) > 0:
                sha.update(fbuf)
                fbuf = f_in.read(HASH_READ_BLOCK)
            content_hash = sha.hexdigest()
        return content_hash

    @functools.lru_cache(maxsize=None)
    def attributes_hash(self):
        return hashlib.sha256(pickle.dumps(self.attributes())).hexdigest()

    def hashes(self):
        return (self.content_hash(), self.attributes_hash())

    def print_summary(self):
        to_print = {
            'Local path': click.format_filename(str(self.path_local_abs)),
            'S3 path': 's3://{0}/{1}'.format(
                self.rules['aws']['s3_bucket_name'], self.s3_path()),
            'Attributes': self.attributes(),
            'Content size': humanize.naturalsize(self.size()),
            'Content hash': self.content_hash(),
            'Attributes hash': self.attributes_hash()
        }
        title = 'File: {0}'.format(click.format_filename(
            str(self.path_local_rel)))
        s3sup.utils.pprint_h3(title)
        s3sup.utils.pprint_dict(to_print)



import re
import toml
import json
import pkgutil

import jsonschema


def directives_for_path(path, rules):
    attrs = {}
    notransfer = {'path', 'path_re', '_comment'}
    try:
        path_specific_rules = rules['path_specific']
    except KeyError:
        return attrs
    for p in path_specific_rules:
        if p['path_re'].match(path) is not None:
            to_add = {k: v for (k, v) in p.items() if k not in notransfer}
            attrs = {**attrs, **to_add}
    return attrs


def _load_schema():
    schema = pkgutil.get_data(__package__, 'rules.schema.json')
    return json.loads(schema)


def _compile_path_regex(rules):
    if 'path_specific' in rules:
        for r in rules['path_specific']:
            r['path_re'] = re.compile(r['path'])
    return rules


def load_rules(rules_path):
    schema = _load_schema()
    with open(rules_path) as rf:
        rules = toml.load(rf)
    jsonschema.validate(rules, schema)
    rules = _compile_path_regex(rules)
    return rules
